/****************************************************
 * ifmon Interface Monitor
 *
 * (c) Hiroaki Hata 2002
 * (c) Miki Takata  2002
 * **************************************************/


#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>     /* the L2 protocols */
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <linux/sockios.h>
#include <linux/if.h>
#include <error.h>
#include <errno.h>
#include <sys/time.h>
#include "ifmon.h"

#define VER  "0.1"
#define BUILT  "Nov.08,2002"

#define IFNAME		"eth0"
#define CFNAME		"ifmon.conf"
#define INTERVAL	10 //sec
#define HASH_SIZE	256
#define MAIL_MAX	16




static struct	ifreq if_rec;
static	int	s;
static	int	timeout;
static	unsigned int cuml;
static	int	mail=0;
static	char	mail_addr[MAIL_MAX][128];
static	char	*maddr[MAIL_MAX];
static	char	from_addr[128];
static	char	smtp_server[1024];
static	int	speed=10; /*10Mbps*/
static	float	limit=10.0;	/*10%*/
static	float   amount=4;      /*4Mbps*/
static	int	dump_mode=0;
static	char	ifname[128];
static	int	no=0;
static	int	verbose=0;

static void recover(void);
static void usage (char *cmd);
static void non_promisc(char *ifname);
static void analyze_paket(int byte,unsigned char *buffer);
static void dump_ether(void );
static void if_mon(void );
static void exiting(int n);
static void credit(void);

typedef struct t_stat {
	int	onoff;
	unsigned char addr[4];
	unsigned long count;
}STAT;

static STAT	stat[HASH_SIZE];

static void credit(void)
{
	printf("-------ifmon interface monitor-----\n");
	printf("version %s\n",VER);
	printf("built   %s\n\n",BUILT);
	printf("(c) Hiroaki Hata \n");
	printf("Network Planning Devision\n");
	printf("NTTPC Communications,Inc.\n");
	printf("----------------------------------\n");
}

static void usage (char *cmd)
{
	fprintf(stderr,"usage:\n");
	fprintf(stderr,"%s     󥿡եeth0Υ˥\n",cmd);
	fprintf(stderr,"%s -r  󥿡եeth0ץߥ⡼ɤ\n",cmd);
}

static void intr_handler(int n)
{
	printf("SIGINTߤդޤ\n");
	recover();
	credit();
	exit(0);
}

static void alarm_handler(int n)
{
//	printf("SIGALRMߤդޤ\n");
	timeout=1;
}

static void non_promisc(char *pifname)
{

	if(pifname ==NULL){
		strcpy(ifname,IFNAME);
	}else if(strlen(pifname)>15 ){
		return;	
	}else{
		strcpy(ifname,pifname);
	}
	
	//åȤ
	s=socket(PF_PACKET,SOCK_RAW,htons(ETH_P_ALL));
	if(s<0){
		sys_err("socket");
		exit(1);
	}
	printf("socketޤ\n");
	//ץߥ⡼ɤѹ
	//󥿥եξ֤
	strcpy(if_rec.ifr_name,ifname);
	if( ioctl(s,SIOCGIFFLAGS,&if_rec) < 0 ){
		//Error
		sys_err("ioctl ּ");
		close(s);
		exit(1);
	}
	//ե饰ѹ
	if_rec.ifr_flags = if_rec.ifr_flags & ~IFF_PROMISC;
	if( ioctl(s,SIOCSIFFLAGS,&if_rec) < 0){
		//Error
		sys_err("ioctl ");
		close(s);
		exit(1);
	}
	printf("ǥХ %s ץߥ⡼ɤ촰λ\n",ifname);
	close(s);
	exit(0);

}

static void recover(void)
{

	//ե饰ѹ
	if_rec.ifr_flags = if_rec.ifr_flags & ~IFF_PROMISC;
	if( ioctl(s,SIOCSIFFLAGS,&if_rec) < 0){
		//Error
		sys_err("ioctl ");
	}
	printf("ǥХ %s ץߥ⡼ɤ촰λ\n",ifname);
	close(s);
}


static void analyze_paket(int byte,unsigned char *buffer)
{
	static int n=1;
	void	*ptr;
	unsigned char frm[2];
	unsigned char addr[4];
	int	ipf=0;
	int	i;
	unsigned int hash;
	int line;
	
	cuml+=byte;
	no++;
	if(dump_mode==1){
		dump_packet(buffer,byte);
		return;
	}
	/*ե졼ॿפ򸡺*/
	ptr=buffer+12;
	memcpy(frm,ptr,2);
	if(frm[0]==0x08 && frm[1]==0x00){
		/*IPv4ե졼Ǥ*/
		ipf=1;
		/*ȯɥ쥹Ф*/
		ptr+=14;
		memcpy(addr,ptr,4);
		/*ץơ֥Ͽ*/
		/*ϥå׻*/
		hash=addr[0]+addr[1]+addr[2]+addr[3];
		hash %=HASH_SIZE;
		/*ơ֥ζ򸡺*/
		line=hash;
		for(i=0;i<HASH_SIZE;i++,line++){
			line %= HASH_SIZE;
			if(stat[line].onoff==0){
				/*򸫤ĤΤǡΥɥ쥹Ͽ*/
				stat[line].onoff=1;
				memcpy(stat[line].addr,addr,4);
				stat[line].count++;
				break;
			}else{
				/*֤ϻ*/
				if(memcmp(stat[line].addr,addr,4)==0){
					/*Υɥ쥹Ͽ*/
					/*ե졼򥫥ȥå*/
					stat[line].count++;
					break;
				}else{
					/*¾ɥ쥹ǻ*/
					continue;
				}
			}
		}/*Υ르ꥺϡϥå奵ˤϳ
		   ǽ⤢뤬®ٸΨŻ뤷פθ̩ڻ뤷*/
	}

	if(verbose==1){
		if(ipf==1){
			printf("%d:%d:%d:%d\t(%02X%02X) %d.%d.%d.%d\n",
					n++,no,byte,cuml,
					frm[0],frm[1],
					addr[0],addr[1],addr[2],addr[3]	
					);
		}else{
			printf("%d:%d:%d:%d\t(%02X%02X)\n",
					n++,no,byte,cuml,frm[0],frm[1]);
		}
	}

}

static void cycle_process(void)
{
	time_t	t;
	long	rate,util;
	char	buff[2048];
	char	buff_1[128];
	char	buff_2[128];
	int i;
	int f;
	static int mcount=0;

	if(mcount!=0) mcount--;

	time(&t);
	/*LANθѥ롼ץåȤ׻*/
	rate=cuml*8/INTERVAL; /*bps*/
	if(verbose==1){
		fprintf(stderr,"******%4d kbps:%d packets\n",rate/1000,no);
	}else{
		/*LANλΨ׻*/
		util=100*rate/(speed*1000000); /*speed=Mbps*/
		if(util>100) util=100;
		sprintf(buff,"%10dkbps:%3d%%:",rate/1000,util);
		for(i=0;i<util/2;i++){
			strcat(buff,"*");
		}
		printf("%s\n",buff);
	}
	/*׾򸡺*/
	/*ߤΥȥեåȥեåĶᤷƤз׻*/
	if(rate>(amount*1000000)){
		sprintf(buff,"Subject: **IFMON alerts**\n");
		if(maddr[0] != NULL){
			sprintf(buff_1,"To: %s",maddr[0]);
			for(i=1;i<10 && maddr[i]!=NULL;i++){
				sprintf(buff_2,", %s",maddr[i]);
				strcat(buff_1,buff_2);
			}
			sprintf(buff_2, "\n");
			strcat(buff_1,buff_2);
			strcat(buff,buff_1);
		}
		if(from_addr[0] != '\0'){
			sprintf(buff_1,"From: %s\n",from_addr);
			strcat(buff,buff_1);
		}
		sprintf(buff_1, "\n");
		strcat(buff,buff_1);
		
		sprintf(buff_1,"Traffic:%dkbps\n",rate/1000);
		strcat(buff,buff_1);
		sprintf(buff_1,"Frames :%d\n",no);
		strcat(buff,buff_1);
		sprintf(buff_1,"------------------\n");
		strcat(buff,buff_1);
		f=0;
		for(i=0;i<HASH_SIZE;i++){
			if(stat[i].onoff==1&& limit<=stat[i].count*100/no){
				sprintf(buff_1,"%d.%d.%d.%d:\t%d:%d%%\n",
						stat[i].addr[0],
						stat[i].addr[1],
						stat[i].addr[2],
						stat[i].addr[3],
						stat[i].count,
						stat[i].count*100/no);
				if(strlen(buff_1)+strlen(buff)<1500){
					strcat(buff,buff_1);
					f=1;
				}
			}
		}
		if(f==1){/*ɬפХݡȯ*/
			printf("%s",buff);
			/*᡼1֤˰*/
			/*mcount0ʤ᡼ФƤ褤*/
			if(mail==1 && mcount==0){
				mcount=3600/INTERVAL;
				SendMail(buff,maddr,from_addr,smtp_server,0);
			}
		}
	}
	cuml=0;
	no=0;
	memset(stat,0,sizeof(stat));
}


static void dump_ether(void )
{
#define BUFFER_SIZE	2048

	unsigned char buffer[BUFFER_SIZE];
	int	byte,n;
	fd_set	fdset;
	struct timeval tv;

	alarm(INTERVAL);
	cuml=0;

	for(;;){
		FD_ZERO(&fdset);
		FD_SET(s,&fdset);
		tv.tv_sec=0;
		tv.tv_usec=500000;/*0.5òʤ*/
		n=select(32,&fdset,NULL,NULL,&tv);
		if(n<0){
			if(errno==EINTR){
			}else{
				sys_err("select");
			}
		}else if(n>0){

			byte=read(s,buffer,BUFFER_SIZE);
			if(byte<=0){
				sys_err("read");
				break;
			}
			analyze_paket(byte,buffer);
		}
		if(timeout==1){
			//ޡե饰ꥻå
			timeout=0;
			//ư
			cycle_process();
			//ޡƥ
			alarm(INTERVAL);
		}
	}
}

static void if_mon(void)
{
	
	//åȤ
	s=socket(PF_PACKET,SOCK_RAW,htons(ETH_P_ALL));
	if(s<0){
		sys_err("socket");
		exit(1);
	}
	printf("socketޤ\n");
	//ץߥ⡼ɤѹ
	//󥿥եξ֤
	strcpy(if_rec.ifr_name,ifname);
	if( ioctl(s,SIOCGIFMTU,&if_rec) < 0 ){
		//Error
		sys_err("ioctl MTU");
		close(s);
		exit(1);
	}
	printf("MTU:%d\n",if_rec.ifr_mtu);
	printf("-------------------------\n");
	if(if_rec.ifr_mtu>2000){
		fprintf(stderr,"MTU2000ХȰѤƤ");
		close(s);
		exit(1);
	}

	if( ioctl(s,SIOCGIFFLAGS,&if_rec) < 0 ){
		//Error
		sys_err("ioctl ּ");
		close(s);
		exit(1);
	}
	//ե饰ѹ
	if_rec.ifr_flags = if_rec.ifr_flags | IFF_PROMISC;
	if( ioctl(s,SIOCSIFFLAGS,&if_rec) < 0){
		//Error
		sys_err("ioctl ");
		close(s);
		exit(1);
	}
	printf("ǥХ %s ץߥ⡼ɤذܹԴλ\n",ifname);
	signal(SIGINT,intr_handler);
	signal(SIGALRM,alarm_handler);
	dump_ether();
	intr_handler(0);
}

static void initialize_param(char *pfname)
{
	char	fname[1024];
	char	buff[1024];
	char	p1[128];
	char	p2[128];
	FILE	*fp;
	int n;
	int m;
	int mc;


	if(pfname==NULL){
		strcpy(fname,CFNAME);
	}else if(strlen(pfname)>512){
		fprintf(stderr,"Filename(%s) too long\n",fname);
		exit(1);
	}else {
		strcpy(fname,pfname);
	}
	fp=fopen(fname,"r");
	if(fp==NULL){
		fprintf(stderr,"File(%s) cannot open\n",fname);
		exit(1);
	}
	m=0;
	mc=0;
	for(;;){
		if(NULL==fgets(buff,128,fp)) 
			break;
		if(*buff=='#'|| *buff=='\0' || *buff=='\r' || *buff=='\n') 
			continue;
		n=sscanf(buff,"%s%s",p1,p2);
		if(strcmp(p1,"IF")==0 && n>=2){
			strcpy(ifname,p2);
			continue;
		}
		if(strcmp(p1,"TO")==0 && n>=2){
			if(mc<MAIL_MAX-1){
				strcpy(mail_addr[mc],p2);
				maddr[mc]=mail_addr[mc];
				mc++;
			}
			m |= 1;
			continue;
		}
		if(strcmp(p1,"FROM")==0 && n>=2){
			strcpy(from_addr,p2);
			m |= 2;
			continue;
		}
		if(strcmp(p1,"SMTP")==0 && n>=2){
			strcpy(smtp_server,p2);
			m |= 4;
			continue;
		}
		if(strcmp(p1,"SPEED")==0 && n>=2){
			speed=atoi(p2);	
			continue;
		}
		if(strcmp(p1,"LIMIT")==0 && n>=2){
			sscanf(p2,"%f",&limit);	
			continue;
		}
		if(strcmp(p1,"TRAFFIC")==0 && n>=2){
			sscanf(p2,"%f",&amount);	
			continue;
		}
		if(strcmp(p1,"DUMP")==0 ){
			dump_mode=1;	
			continue;
		}
		if(strcmp(p1,"VERBOSE")==0 ){
			verbose=1;	
			continue;
		}
		fprintf(stderr,"Err:%s\n",buff);
		exit(1);
	}
	if(m==7) mail=1;
	printf("Paramters\n");
	printf("ifname=%s\n",ifname);
	printf("speed =%d\n",speed);
	printf("limit =%f\n",limit);
	printf("amount=%f\n",amount);
	printf("dump_mode=%d\n",dump_mode);
	printf("mail=%d\n",mail);
	if(mail==1){
		for(n=0;n<MAIL_MAX;n++){
			if(maddr[n]==NULL)break;
			printf("to     =%s\n",maddr[n]);
		}
		printf("from   =%s\n",from_addr);
		printf("smtp   =%s\n",smtp_server);
	}
	printf("-------------------------\n");
	return;
}



int main(int argc,char **argv)
{
	int	i;
	char	text[128];


	credit();
	for(i=0;i<MAIL_MAX;i++){
		maddr[i]=NULL;
		mail_addr[i][0]='\0';
	}

	memset(stat,0,sizeof(stat));

	timeout=0;
	strcpy(ifname,IFNAME);

	if(argc == 1){
		initialize_param(NULL);
		if_mon();
		exit(1);
	} else if(strcmp(argv[1],"-f")==0){
		if(argc==3){
			initialize_param(argv[2]);
			if_mon();
			exit(0);
		}else{
			usage(argv[0]);
		}
	} else if(strcmp(argv[1],"-r")==0){
		if(argc==3){
			non_promisc(argv[2]);
		}else{
			non_promisc(NULL);
		}
		exit(0);
	} else if(strcmp(argv[1],"h")==0 || strcmp(argv[1],"-h")==0){
		usage(argv[0]);
		exit(0);
	}
	usage(argv[0]);
	return 0;

}

